UpvarCapture

Enum UpvarCapture 

pub enum UpvarCapture {
    ByValue,
    ByUse,
    ByRef(BorrowKind),
}
Information describing the capture of an upvar. This is computed during typeck, specifically by regionck.

Variants



ByValue

Upvar is captured by value. This is always true when the closure is labeled move, but can also be true in other cases depending on inference.



ByUse

Upvar is captured by use. This is true when the closure is labeled use.



ByRef(BorrowKind)

Upvar is captured by reference.