pub struct CapturedPlace {
pub place: Place,
pub info: CaptureInfo,
pub mutability: Mutability,
}Expand description
A composite describing a Place that is captured by a closure.
Fields
place: PlaceThe Place that is captured.
info: CaptureInfoCaptureKind and expression(s) that resulted in such capture of place.
mutability: MutabilityRepresents if place can be mutated or not.
Implementations
Sourceimpl CapturedPlace
impl CapturedPlace
pub fn is_by_ref(&self) -> bool
pub fn captured_local(&self) -> BindingId
Sourcepub fn captured_ty<'db>(&self, db: &'db dyn HirDatabase) -> Ty<'db>
pub fn captured_ty<'db>(&self, db: &'db dyn HirDatabase) -> Ty<'db>
The type of the capture stored in the closure, which is different from the type of the captured place if we capture by reference.