pub struct PatAdjustment {
pub kind: PatAdjust,
pub source: StoredTy,
}Expand description
Represents an implicit coercion applied to the scrutinee of a match before testing a pattern against it. Currently, this is used only for implicit dereferences.
Fields
kind: PatAdjustsource: StoredTyThe type of the scrutinee before the adjustment is applied, or the “adjusted type” of the pattern.