pub struct ClosureData {
pub min_captures: FxIndexMap<BindingId, Vec<CapturedPlace>>,
pub fake_reads: Box<[(Place, FakeReadCause, SmallVec<[CaptureSourceStack; 2]>)]>,
}Fields
min_captures: FxIndexMap<BindingId, Vec<CapturedPlace>>Tracks the minimum captures required for a closure;
see MinCaptureInformationMap for more details.
fake_reads: Box<[(Place, FakeReadCause, SmallVec<[CaptureSourceStack; 2]>)]>Tracks the fake reads required for a closure and the reason for the fake read. When performing pattern matching for closures, there are times we don’t end up reading places that are mentioned in a closure (because of _ patterns). However, to ensure the places are initialized, we introduce fake reads. Consider these two examples:
let x: u8;
let c = || match x { _ => () };In this example, we don’t need to actually read/borrow x in c, and so we don’t
want to capture it. However, we do still want an error here, because x should have
to be initialized at the point where c is created. Therefore, we add a “fake read”
instead.
let c = || {
let (t1, t2) = t;
}In the second example, we capture the disjoint fields of t (t.0 & t.1), but
we never capture t. This becomes an issue when we build MIR as we require
information on t in order to create place t.0 and t.1. We can solve this
issue by fake reading t.