pub(crate) type RootVariableMinCaptureList = FxIndexMap<BindingId, Vec<CapturedPlace>>;Expand description
Part of MinCaptureInformationMap; Maps a root variable to the list of CapturedPlace.
Used to track the minimum set of Places that need to be captured to support all
Places captured by the closure starting at a given root variable.
This provides a convenient and quick way of checking if a variable being used within a closure is a capture of a local variable.
Aliased Type
pub(crate) struct RootVariableMinCaptureList { /* private fields */ }